@@ -21,6 +21,7 @@ import copy
 import datetime
 import errno
 import functools
+import locale
 import os
 import re
 import shutil
@@ -193,8 +194,9 @@ def _files_same(files, regexes, comparison_args):
         # For probable text files, discard newline differences to match
         # file comparisons.
         contents = ["\n".join(c.splitlines()) for c in contents]
-        for r in regexes:
-            contents = [re.sub(r, "", c) for c in contents]
+
+        contents = [misc.apply_text_filters(c, regexes) for c in contents]
+
         if ignore_blank_lines:
             contents = [remove_blank_lines(c) for c in contents]
         result = SameFiltered if all_same(contents) else Different
@@ -674,6 +676,8 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         todo = [rootpath]
         expanded = set()
 
+        tuple_tree_path = lambda p: tuple(p.get_indices())
+
         shadowed_entries = []
         invalid_filenames = []
         while len(todo):
@@ -783,7 +787,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                         not all(os.path.isdir(f) for f in roots)):
                     self.model.add_empty(it)
                     if self.model.iter_parent(it) is None:
-                        expanded.add(rootpath)
+                        expanded.add(tuple_tree_path(rootpath))
                 else:
                     # At this point, we have an empty folder tree node; we can
                     # prune this and any ancestors that then end up empty.
@@ -794,7 +798,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                         # no siblings. If we're here, we have an empty tree.
                         if parent is None:
                             self.model.add_empty(it)
-                            expanded.add(rootpath)
+                            expanded.add(tuple_tree_path(rootpath))
                             break
 
                         # Remove the current row, and then revalidate all
@@ -810,7 +814,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                         it = parent
 
             if differences:
-                expanded.add(path)
+                expanded.add(tuple_tree_path(path))
 
         if invalid_filenames or shadowed_entries:
             self._show_tree_wide_errors(invalid_filenames, shadowed_entries)
@@ -819,7 +823,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
 
         self.treeview[0].expand_to_path(Gtk.TreePath(("0",)))
         for path in sorted(expanded):
-            self.treeview[0].expand_to_path(path)
+            self.treeview[0].expand_to_path(Gtk.TreePath(path))
         yield _("[%s] Done") % self.label_text
 
         self.scheduler.add_task(self.on_treeview_cursor_changed)
@@ -852,7 +856,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
 
         for pane in range(self.num_panes):
             msgarea = self.msgarea_mgr[pane].new_from_text_and_icon(
-                Gtk.STOCK_INFO, primary, secondary)
+                'dialog-information-symbolic', primary, secondary)
             button = msgarea.add_button(_("Hide"), Gtk.ResponseType.CLOSE)
             if pane == 0:
                 button.props.label = _("Hi_de")
@@ -902,17 +906,8 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                 else:
                     continue
                 secondary = "\n".join(messages)
-                self.add_dismissable_msg(pane, Gtk.STOCK_DIALOG_ERROR, header,
-                                         secondary)
-
-    def add_dismissable_msg(self, pane, icon, primary, secondary):
-        msgarea = self.msgarea_mgr[pane].new_from_text_and_icon(
-                        icon, primary, secondary)
-        msgarea.add_button(_("Hi_de"), Gtk.ResponseType.CLOSE)
-        msgarea.connect("response",
-                        lambda *args: self.msgarea_mgr[pane].clear())
-        msgarea.show_all()
-        return msgarea
+                self.msgarea_mgr[pane].add_dismissable_msg(
+                    'dialog-error-symbolic', header, secondary)
 
     def copy_selected(self, direction):
         assert direction in (-1, 1)
@@ -1325,6 +1320,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             all_present_same = self.file_compare(lof, regexes)
         different = 1
         one_isdir = [None for i in range(self.model.ntree)]
+        locale_encoding = locale.getpreferredencoding()
         for j in range(self.model.ntree):
             if mod_times[j]:
                 isdir = os.path.isdir( files[j] )
@@ -1352,6 +1348,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                 TIME = self.model.column_index(COL_TIME, j)
                 mod_datetime = datetime.datetime.fromtimestamp(mod_times[j])
                 time_str = mod_datetime.strftime("%a %d %b %Y %H:%M:%S")
+                time_str = time_str.decode(locale_encoding, errors='replace')
                 self.model.set_value(it, TIME, time_str)
 
                 def natural_size(bytes):
@@ -1571,7 +1568,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
     def on_refresh_activate(self, *extra):
         self.on_fileentry_file_set(None)
 
-    def on_delete_event(self, appquit=0):
+    def on_delete_event(self):
         for h in self.settings_handlers:
             meldsettings.disconnect(h)
         self.emit('close', 0)
