@@ -37,11 +37,12 @@ class MeldBuffer(GtkSource.Buffer):
         ('highlight-syntax', 'highlight-syntax'),
     )
 
-    def __init__(self, filename=None):
+    def __init__(self):
         GtkSource.Buffer.__init__(self)
         bind_settings(self)
-        self.data = MeldBufferData(filename)
+        self.data = MeldBufferData()
         self.user_action_count = 0
+        self.undo_sequence = None
         meldsettings.connect('changed', self.on_setting_changed)
         self.set_style_scheme(meldsettings.style_scheme)
 
@@ -51,8 +52,12 @@ class MeldBuffer(GtkSource.Buffer):
 
     def do_begin_user_action(self, *args):
         self.user_action_count += 1
+        if self.undo_sequence:
+            self.undo_sequence.begin_group()
 
     def do_end_user_action(self, *args):
+        if self.undo_sequence:
+            self.undo_sequence.end_group()
         self.user_action_count -= 1
 
     def do_apply_tag(self, tag, start, end):
@@ -68,14 +73,6 @@ class MeldBuffer(GtkSource.Buffer):
             return
         return GtkSource.Buffer.do_apply_tag(self, tag, start, end)
 
-    def reset_buffer(self, filename):
-        """Clear the contents of the buffer and reset its metadata"""
-        self.delete(*self.get_bounds())
-        label = self.data.label if self.data.filename == filename else filename
-        self.data.reset()
-        self.data.filename = filename
-        self.data.label = label
-
     def get_iter_at_line_or_eof(self, line):
         """Return a Gtk.TextIter at the given line, or the end of the buffer.
 
@@ -111,23 +108,20 @@ class MeldBufferData(GObject.GObject):
         str('file-changed'): (GObject.SignalFlags.RUN_FIRST, None, ()),
     }
 
-    def __init__(self, filename=None):
+    def __init__(self):
         GObject.GObject.__init__(self)
-        self.reset()
-        self.label = self.filename = filename
-
-    def reset(self):
-        self.modified = False
-        self.writable = True
-        self.editable = True
-        self._monitor = None
-        self._mtime = None
-        self._disk_mtime = None
-        self.filename = None
-        self.savefile = None
+        self._gfile = None
         self._label = None
-        self.encoding = None
-        self.newlines = None
+        self._monitor = None
+        self._sourcefile = None
+        self.reset(gfile=None)
+
+    def reset(self, gfile):
+        same_file = gfile and self._gfile and gfile.equal(self._gfile)
+        self.gfile = gfile
+        self.label = self._label if same_file else self.filename
+        self.loaded = False
+        self.savefile = None
 
     def __del__(self):
         self._disconnect_monitor()
@@ -145,17 +139,19 @@ class MeldBufferData(GObject.GObject):
         self._label = fallback_decode(value, encodings, lossy=True)
 
     def _connect_monitor(self):
-        if self._filename:
-            monitor = Gio.File.new_for_path(self._filename).monitor_file(
-                Gio.FileMonitorFlags.NONE, None)
-            handler_id = monitor.connect('changed', self._handle_file_change)
-            self._monitor = monitor, handler_id
+        if not self._gfile:
+            return
+        monitor = self._gfile.monitor_file(Gio.FileMonitorFlags.NONE, None)
+        handler_id = monitor.connect('changed', self._handle_file_change)
+        self._monitor = monitor, handler_id
 
     def _disconnect_monitor(self):
-        if self._monitor:
-            monitor, handler_id = self._monitor
-            monitor.disconnect(handler_id)
-            monitor.cancel()
+        if not self._monitor:
+            return
+        monitor, handler_id = self._monitor
+        monitor.disconnect(handler_id)
+        monitor.cancel()
+        self._monitor = None
 
     def _query_mtime(self, gfile):
         try:
@@ -174,20 +170,53 @@ class MeldBufferData(GObject.GObject):
         self._disk_mtime = mtime
 
     @property
-    def filename(self):
-        return self._filename
+    def gfile(self):
+        return self._gfile
 
-    @filename.setter
-    def filename(self, value):
+    @gfile.setter
+    def gfile(self, value):
         self._disconnect_monitor()
-        self._filename = value
+        self._gfile = value
+        self._sourcefile = GtkSource.File()
+        self._sourcefile.set_location(value)
+
+        # This is aiming to maintain existing behaviour for filename. The
+        # behaviour is however wrong and should be fixed.
+        self.filename = value.get_path().decode('utf8') if value else None
         self.update_mtime()
         self._connect_monitor()
 
+    @property
+    def sourcefile(self):
+        return self._sourcefile
+
+    @property
+    def gfiletarget(self):
+        if self.savefile:
+            return Gio.File.new_for_path(self.savefile)
+        return self.gfile
+
+    @property
+    def is_special(self):
+        try:
+            info = self._gfile.query_info(
+                Gio.FILE_ATTRIBUTE_STANDARD_TYPE, 0, None)
+            return info.get_file_type() == Gio.FileType.SPECIAL
+        except (AttributeError, GLib.GError):
+            return False
+
+    @property
+    def writable(self):
+        try:
+            info = self.gfiletarget.query_info(
+                Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE, 0, None)
+        except (AttributeError, GLib.GError):
+            return False
+        return info.get_attribute_boolean(Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE)
+
     def update_mtime(self):
-        if self._filename:
-            gfile = Gio.File.new_for_path(self._filename)
-            self._disk_mtime = self._query_mtime(gfile)
+        if self._gfile:
+            self._disk_mtime = self._query_mtime(self._gfile)
             self._mtime = self._disk_mtime
 
     def current_on_disk(self):
@@ -288,10 +317,14 @@ class BufferAction(object):
         start = self.buffer.get_iter_at_offset(self.offset)
         end = self.buffer.get_iter_at_offset(self.offset + len(self.text))
         self.buffer.delete(start, end)
+        self.buffer.place_cursor(end)
+        return [self]
 
     def insert(self):
         start = self.buffer.get_iter_at_offset(self.offset)
+        self.buffer.place_cursor(start)
         self.buffer.insert(start, self.text)
+        return [self]
 
 
 class BufferInsertionAction(BufferAction):
